#include "Scene.hxx"
#include "PerspectiveCamera.hxx"
#include <limits>

Scene::Scene()
  : m_camera(new PerspectiveCamera(Vec3f(0,0,8),
				   Vec3f(0,0,-1),
				   Vec3f(0,1,0),
				   50,
				   640,
				   480)
	     ),
    m_bgColor(Vec3f(0,0,0))
{
}

Scene::~Scene()
{
  delete m_camera;
}

Scene::Scene(const Scene& s)
{
  operator=(s);
}

Scene&
Scene::operator=(const Scene& s)
{
  return *this;
}

void
Scene::Add(Primitive* p)
{
  m_primitives.push_back(p);
}


bool
Scene::Intersect(Ray& ray)
{
  bool hit = false;
  float t = std::numeric_limits<float>::max();
  for (std::vector<Primitive*>::iterator i = m_primitives.begin();
    i != m_primitives.end(); i++) {
    // store closest object hit
    if (hit |= (*i)->Intersect(ray)) {
      if (ray.t() < t) {
        ray.setHit(*i);
        t = ray.t();
      }
    }
  }
  return hit;
}

bool
Scene::Occluded(Ray& ray)
{
  return false;
}

Vec3f
Scene::RayTrace(Ray& ray)
{
  if (Intersect(ray)) {
    return ray.hit()->shader()->Shade(ray);
  } else {
    return Vec3f(0,0,0);
  }
}

const Camera*
Scene::camera() const
{
  return m_camera;
}
